import bs
import bsUtils
import bsSpaz
import bsFlag
import bsMap
import bsAnnounce
import bsLobby
import random
import bsPowerup
import bsScoreBoard
import bsScoreSet
import bsBomb
import weakref
import copy
import bsUI
import bsAchievement

# team info
gTeamColors = ((0.1,0.25,1.0),
               (1.0,0.25,0.2))
gTeamNames = ("Blue","Red")

class Results(object):
    """Standard results for a completed team-game activity."""
    def __init__(self,scoreName='Score',lowerIsBetter=False,noneIsWinner=False,scoreType='points'):
        self._lowerIsBetter = lowerIsBetter
        self._scores = {}
        self._scoreName = scoreName
        self._noneIsWinner = noneIsWinner
        self._scoreType = scoreType
    def setTeamScore(self,team,score):
        self._scores[team.teamID] = (weakref.ref(team),score)
    def getTeamScore(self,team):
        for score in self._scores.values():
            if score[0]() is team:
                return score[1]
        raise Exception("team not found in results")
    def getTeams(self):
        'return valid teams'
        teams = []
        for scoreInfo in self._scores.values():
            team = scoreInfo[0]()
            if team is not None:
                teams.append(team)
        return teams
    def hasScoreForTeam(self,team):
        for score in self._scores.values():
            if score[0]() is team:
                return True
        return False
    def getTeamScoreStr(self,team):
        for score in self._scores.values():
            if score[0]() is team:
                if score[1] is None: return '-'
                if self._scoreType == 'seconds': return bsUtils.getTimeString(score[1]*1000,centi=False)
                elif self._scoreType == 'milliseconds': return bsUtils.getTimeString(score[1],centi=True)
                else: return str(score[1])
        raise Exception("team not found in results")
    def getScoreName(self):
        return self._scoreName
    def getLowerIsBetter(self):
        return self._lowerIsBetter
    def getWinningTeam(self):
        winners = self.getWinners()
        if len(winners) > 0 and len(winners[0]['teams']) == 1: return winners[0]['teams'][0]
        else: return None

    def getWinners(self):
        'Return an ordered list of dicts containing score and teams'
        winners = {}
        # filter out any dead teams
        scores = [score for score in self._scores.values() if score[0]() is not None and score[1] is not None]
        for score in scores:
            try: s = winners[score[1]]
            except Exception: s = winners[score[1]] = []
            s.append(score[0]())
        results = winners.items()
        results.sort(reverse=not self._lowerIsBetter)


        # tack a group with all our 'None' scores onto the end
        noneTeams = [score[0]() for score in self._scores.values() if score[0]() is not None and score[1] is None]
        if len(noneTeams) > 0:
            nones = [(None,noneTeams)]
            if self._noneIsWinner: results = nones + results
            else: results = results + nones
        return [{'score':r[0],'teams':r[1]} for r in results]


class ShuffleList(object):
    """ shuffles a set of games with some smarts to avoid repeats in maps or game types """
    def __init__(self,items,shuffle=True):
        self.sourceList = items
        self.shuffle = shuffle
        self.shuffleList = []
        self.lastGotten = None

    def pullNext(self):

        # refill our list if its empty
        if len(self.shuffleList) == 0:
            self.shuffleList = list(self.sourceList)

        # ok now find an index we should pull
        index = 0

        if self.shuffle:
            for i in range(4):
                index = random.randrange(0,len(self.shuffleList))
                testObj = self.shuffleList[index]
                # if the new one is the same map or game-type as the previous, lets try to keep looking..
                if len(self.shuffleList) > 1 and self.lastGotten != None:
                    if testObj['map'] == self.lastGotten['map']:
                        continue
                    if testObj['type'] == self.lastGotten['type']:
                        continue
                # sufficiently different.. lets go with it
                break

        obj = self.shuffleList.pop(index)
        self.lastGotten = obj
        return obj



class TeamsScoreScreenActivity(bs.ScoreScreenActivity):

    def __init__(self,session,settings={}):
        bs.ScoreScreenActivity.__init__(self,session,settings=settings)
        self._scoreDisplaySound = bs.getSound("scoreHit01")
        self._scoreDisplaySoundSmall = bs.getSound("scoreHit02")

    def onBegin(self,showUpNext=True,customContinueMessage=None):
        bs.ScoreScreenActivity.onBegin(self,customContinueMessage=customContinueMessage)
        if showUpNext:
            bsAnnounce.Text(self,bs.Actor.RETAINED,
                            (bs.getResource('upNextText').replace('${COUNT}',str(self.getSession().getGameNumber()+1))
                             +'   '+self.getSession().getNextGameName()),
                            maxWidth=900,
                            hAttach='center',
                            vAttach='bottom',
                            hAlign='center',
                            vAlign='center',
                            position=(0,53),
                            flash=False,
                            color=(0.3,0.3,0.3,1.0),
                            transition='fadeIn',
                            transitionDelay=2000)

    def showPlayerScores(self,delay=2500,results=None,scale=1.0,xOffset=0,yOffset=0):
        tsVOffset = 150.0+yOffset
        tsHOffs = 80.0+xOffset
        tDelay = delay
        spacing = 40

        isFreeForAll = self.getSession().getSessionType() == 'freeForAll'

        def _getPlayerScore(player):
            if isFreeForAll and results is not None: return results.getTeamScore(player.getTeam())
            else: return player.accumScore

        def _getPlayerScoreStr(player):
            if isFreeForAll and results is not None: return results.getTeamScoreStr(player.getTeam())
            else: return str(player.accumScore)

        # getValidPlayers() can return players that are no longer in the game.. if we're using results we have to filter
        # those out (since they're not in results and that's where we pull their scores from)
        if results is not None:
            playersSorted = []
            validPlayers = self.scoreSet.getValidPlayers().items()

            def _getPlayerScoreSetEntry(player):
                for p in validPlayers:
                    if p[1].getPlayer() is player: return p[1]
                return None

            # results is already sorted; just convert it into a list of score-set-entries
            for winner in results.getWinners():
                for team in winner['teams']:
                    if len(team.players) == 1:
                        playerEntry = _getPlayerScoreSetEntry(team.players.values()[0])
                        if playerEntry is not None:
                            playersSorted.append(playerEntry)
        else:
            playersSorted = [[_getPlayerScore(p),name,p] for name,p in self.scoreSet.getValidPlayers().items()]
            playersSorted.sort(reverse=(results is None or not results.getLowerIsBetter()))
            playersSorted = [p[2] for p in playersSorted if p[2]] # just want living player entries


        vOffs = -140.0 + spacing*len(playersSorted)*0.5

        def _txt(xOffs,yOffs,text,hAlign='right',extraScale=1.0,maxWidth=120.0):
            bsAnnounce.Text(self,bs.Actor.RETAINED,
                            text,color=(0.5,0.5,0.5,0.5),
                            position=(tsHOffs+xOffs*scale,tsVOffset+(vOffs+yOffs+4.0)*scale),
                            hAlign=hAlign,vAlign='center',scale=0.8*scale*extraScale,maxWidth=maxWidth,transition='inLeft',transitionDelay=tDelay)

        #_txt(180,38,'Game '+str(self.getSession().getGameNumber())+' Leaders',hAlign='center',extraScale=1.4,maxWidth=None)
        _txt(180,38,bs.getResource('gameLeadersText').replace('${COUNT}',str(self.getSession().getGameNumber())),hAlign='center',extraScale=1.4,maxWidth=None)
        _txt(-15,0,bs.getResource('playerText'),hAlign='left')
        _txt(180,0,bs.getResource('killsText'))
        _txt(265,0,bs.getResource('deathsText'))

        scoreName = 'Score' if results is None else results.getScoreName()
        try: translated = bs.getResource('translations.scoreNames')[scoreName]
        except Exception: translated = None
        if translated is None: translated = scoreName
        _txt(380,0,translated)

        topKillCount = 0
        topKilledCount = 99999
        topScore = 0 if (len(playersSorted) == 0) else _getPlayerScore(playersSorted[0])

        #for score,name,p in playersSorted:
        for p in playersSorted:
            topKillCount = max(topKillCount,p.accumKillCount)
            topKilledCount = min(topKilledCount,p.accumKilledCount)

        def _scoreTxt(text,xOffs,highlight,delay,maxWidth=70.0):
            bsAnnounce.Text(self,bs.Actor.RETAINED,
                            text,position=(tsHOffs+xOffs*scale,tsVOffset+(vOffs+15)*scale),
                            scale=scale,color=(1.0,0.7,0.3,1.0) if highlight else (0.5,0.5,0.5,0.5),
                            hAlign='right',vAlign='center',maxWidth=maxWidth,transition='inLeft',transitionDelay=tDelay+delay)
        for p in playersSorted:
            tDelay += 50
            vOffs -= spacing
            bsAnnounce.Image(self,bs.Actor.RETAINED,
                             p.getIcon(),position=(tsHOffs-12*scale,tsVOffset+(vOffs+17)*scale),
                             scale=(30.0*scale,30.0*scale),transition='inLeft',transitionDelay=tDelay)
            bsAnnounce.Text(self,bs.Actor.RETAINED,
                            p.getName(),maxWidth=160,scale=0.75*scale,position=(tsHOffs+10.0*scale,tsVOffset+(vOffs+15)*scale),
                            hAlign='left',vAlign='center',color=bsUtils.getSafeDisplayColor(p.getTeam().color+(1,)),transition='inLeft',
                            transitionDelay=tDelay)
            _scoreTxt(str(p.accumKillCount),180,p.accumKillCount==topKillCount,100)
            _scoreTxt(str(p.accumKilledCount),265,p.accumKilledCount==topKilledCount,100)
            _scoreTxt(_getPlayerScoreStr(p),380,_getPlayerScore(p)==topScore,200)


class FreeForAllVictoryScoreScreenActivity(TeamsScoreScreenActivity):

    def __init__(self,session,settings={}):
        TeamsScoreScreenActivity.__init__(self,session,settings=settings)
        self._transitionTime = 500 # keeps prev activity alive while we fade in

        self._cymbalSound = bs.getSound('cymbal')

    def onBegin(self):
        TeamsScoreScreenActivity.onBegin(self)

        yBase = 100
        tsHOffs = -305
        tDelay = 1000
        scale = 1.2
        spacing = 37.0

        # we include name and previous score in the sort to reduce the amount of random jumping around the list we do in cases of ties
        playerOrderPrev = self.players.values()
        playerOrderPrev.sort(reverse=True,key=lambda player:(player.getTeam().sessionData['previousScore'],player.getName()))
        playerOrder = self.players.values()
        playerOrder.sort(reverse=True,key=lambda player:(player.getTeam().sessionData['score'],player.getTeam().sessionData['score'],player.getName()))

        vOffs = -74.0 + spacing*len(playerOrderPrev)*0.5

        delay1 = 1300+100
        delay2 = 2900+100
        delay3 = 2900+100

        orderChange = playerOrder != playerOrderPrev

        if orderChange:
            delay3 += 1500

        bs.gameTimer(300,bs.WeakCall(bs.playSound,self._scoreDisplaySound))
        self.showPlayerScores(delay=1,results=self.settings['results'],scale=1.2,xOffset=-110)
            
        soundTimes = set()

        def _scoreTxt(text,xOffs,yOffs,highlight,delay,extraScale,flash=False):
            return bsAnnounce.Text(self,bs.Actor.RETAINED,
                                   text,position=(tsHOffs+xOffs*scale,yBase+(yOffs+vOffs+2.0)*scale),scale=scale*extraScale,
                                   color=(1.0,0.7,0.3,1.0) if highlight else (0.7,0.7,0.7,0.7),
                                   hAlign='right',transition='inLeft',transitionDelay=tDelay+delay,flash=flash)
        vOffs -= spacing

        slideAmt = 0.0

        transitionTime = 250
        transitionTime2 = 250

        title = bsAnnounce.Text(self,bs.Actor.RETAINED,
                        bs.getResource('firstToSeriesText').replace('${COUNT}',str(self.getSession()._ffaSeriesLength)),scale=1.05*scale,position=(tsHOffs-0.0*scale,yBase+(vOffs+50.0)*scale),
                        hAlign='center',color=(0.5,0.5,0.5,0.5),transition='inLeft',
                        transitionDelay=tDelay)

        vOffs -= 25
        vOffsStart = vOffs

        bs.gameTimer(tDelay+delay3,bs.WeakCall(self._safeAnimate,title.translateCombine,'input0',{0:tsHOffs-0.0*scale,transitionTime2:tsHOffs-(0.0+slideAmt)*scale}))

        for i,player in enumerate(playerOrderPrev):
            vOffs2 = vOffsStart - spacing * (playerOrder.index(player))

            bs.gameTimer(tDelay+300,bs.WeakCall(bs.playSound,self._scoreDisplaySoundSmall))
            if orderChange:
                bs.gameTimer(tDelay+delay2+100,bs.WeakCall(bs.playSound,self._cymbalSound))

            img = bsAnnounce.Image(self,bs.Actor.RETAINED,
                                   player.getIcon(),position=(tsHOffs-72.0*scale,yBase+(vOffs+17.0)*scale),
                                   scale=(30.0*scale,30.0*scale),transition='inLeft',transitionDelay=tDelay)
            bs.gameTimer(tDelay+delay2,bs.WeakCall(self._safeAnimate,img.translateCombine,'input1',{0:yBase+(vOffs+17.0)*scale, transitionTime:yBase+(vOffs2+17.0)*scale}))
            bs.gameTimer(tDelay+delay3,bs.WeakCall(self._safeAnimate,img.translateCombine,'input0',{0:tsHOffs-72.0*scale, transitionTime2:tsHOffs-(72.0+slideAmt)*scale}))
            txt = bsAnnounce.Text(self,bs.Actor.RETAINED,
                            player.getName(),maxWidth=130.0,scale=0.75*scale,position=(tsHOffs-50.0*scale,yBase+(vOffs+15.0)*scale),
                            hAlign='left',vAlign='center',color=bsUtils.getSafeDisplayColor(player.getTeam().color+(1,)),transition='inLeft',
                            transitionDelay=tDelay)
            bs.gameTimer(tDelay+delay2,bs.WeakCall(self._safeAnimate,txt.translateCombine,'input1',{0:yBase+(vOffs+15.0)*scale,transitionTime:yBase+(vOffs2+15.0)*scale}))
            bs.gameTimer(tDelay+delay3,bs.WeakCall(self._safeAnimate,txt.translateCombine,'input0',{0:tsHOffs-50.0*scale,transitionTime2:tsHOffs-(50.0+slideAmt)*scale}))

            txtNum = bsAnnounce.Text(self,bs.Actor.RETAINED,
                            '#'+str(i+1),scale=0.55*scale,position=(tsHOffs-95.0*scale,yBase+(vOffs+8.0)*scale),
                            hAlign='right',color=(0.6,0.6,0.6,0.6),transition='inLeft',transitionDelay=tDelay)
            bs.gameTimer(tDelay+delay3,bs.WeakCall(self._safeAnimate,txtNum.translateCombine,'input0',{0:tsHOffs-95.0*scale,transitionTime2:tsHOffs-(95.0+slideAmt)*scale}))

            sTxt = _scoreTxt(str(player.getTeam().sessionData['previousScore']),80,0,False,0,1.0)
            bs.gameTimer(tDelay+delay2,bs.WeakCall(self._safeAnimate,sTxt.translateCombine,'input1',{0:yBase+(vOffs+2.0)*scale, transitionTime:yBase+(vOffs2+2.0)*scale}))
            bs.gameTimer(tDelay+delay3,bs.WeakCall(self._safeAnimate,sTxt.translateCombine,'input0',{0:tsHOffs+80*scale, transitionTime2:tsHOffs+(80-slideAmt)*scale}))

            scoreChange = player.getTeam().sessionData['score'] - player.getTeam().sessionData['previousScore']
            if scoreChange > 0:
                x = 113
                y = 3.0
                sTxt2 = _scoreTxt('+'+str(scoreChange),x,y,True,0,0.7,flash=True)
                bs.gameTimer(tDelay+delay2,bs.WeakCall(self._safeAnimate,sTxt2.translateCombine,'input1',{0:yBase+(vOffs+y+2.0)*scale, transitionTime:yBase+(vOffs2+y+2.0)*scale}))
                bs.gameTimer(tDelay+delay3,bs.WeakCall(self._safeAnimate,sTxt2.translateCombine,'input0',{0:tsHOffs+x*scale, transitionTime2:tsHOffs+(x-slideAmt)*scale}))
                
                bs.gameTimer(tDelay+delay1,bs.WeakCall(bs.setAttr,sTxt.node,'color',(1,1,1,1)))
                for j in range(scoreChange):
                    bs.gameTimer(tDelay+delay1+150*j,bs.WeakCall(bs.setAttr,sTxt.node,'text',str(player.getTeam().sessionData['previousScore']+j+1),ignoreMissingNode=True))
                    t = tDelay+delay1+150*j
                    if not t in soundTimes:
                        soundTimes.add(t)
                        bs.gameTimer(t,bs.WeakCall(bs.playSound,self._scoreDisplaySoundSmall))
                    
            vOffs -= spacing

    def _safeAnimate(self,node,attr,keys):
        if bs.nodeExists(node):
            bsUtils.animate(node,attr,keys)


class DrawScoreScreenActivity(TeamsScoreScreenActivity):

    def __init__(self,session,settings={}):
        TeamsScoreScreenActivity.__init__(self,session,settings=settings)

    def onTransitionIn(self):
        TeamsScoreScreenActivity.onTransitionIn(self,music=None)
        
    def onBegin(self):
        TeamsScoreScreenActivity.onBegin(self)
        bsAnnounce.ZoomText(self,bs.Actor.RETAINED,
                            bs.getResource('drawText'),position=(0,0),
                            maxWidth=400,
                            shiftPosition=(-220,0),shiftDelay=2000,
                            flash=False,trail=False,jitter=1.0)
        bs.gameTimer(350,bs.WeakCall(bs.playSound,self._scoreDisplaySound))

        if 'results' in self.settings: r = self.settings['results']
        else: r = None

        self.showPlayerScores(results=r)


class TeamVictoryScoreScreenActivity(TeamsScoreScreenActivity):

    def __init__(self,session,settings={}):
        TeamsScoreScreenActivity.__init__(self,session,settings=settings)

    def onBegin(self):
        TeamsScoreScreenActivity.onBegin(self)

        height = 130
        activeTeamCount = len(self.teams)
        v = (height*activeTeamCount)/2 - height/2
        i = 0
        shiftTime = 2500
        bsAnnounce.ZoomText(self,bs.Actor.RETAINED,
                            bs.getResource('bestOfSeriesText').replace('${COUNT}',str(self.getSession()._seriesLength)),
                            position=(0,175),
                            shiftPosition=(-250,175),shiftDelay=2500,
                            flash=False,trail=False,hAlign='center',
                            scale=0.25,
                            color=(0.5,0.5,0.5,1.0),jitter=3.0)
        for team in self.teams.values():
            bs.gameTimer(i*150+150,bs.WeakCall(self._showTeamName,v-i*height,team,i*200,shiftTime-(i*150+150)))
            bs.gameTimer(i*150+500,bs.WeakCall(bs.playSound,self._scoreDisplaySoundSmall))
            scored = (team is self.settings['winner'])
            delay = 200
            if scored:
                delay = 1200
                bs.gameTimer(i*150+200,bs.WeakCall(self._showTeamOldScore,v-i*height,team,shiftTime-(i*150+200)))
                bs.gameTimer(i*150+1500,bs.WeakCall(bs.playSound,self._scoreDisplaySound))

            bs.gameTimer(i*150+delay,bs.WeakCall(self._showTeamScore,v-i*height,team,scored,i*200+100,shiftTime-(i*150+delay)))
            i += 1
        self.showPlayerScores()

    def _showTeamName(self,posV,team,killDelay,shiftDelay):
        bsAnnounce.ZoomText(self,bs.Actor.RETAINED,
                            team.name+":",position=(100,posV),
                            shiftPosition=(-150,posV),shiftDelay=shiftDelay,
                            flash=False,trail=False,hAlign='right',maxWidth=300,
                            color=team.color,jitter=1.0)
    def _showTeamOldScore(self,posV,team,shiftDelay):
        bsAnnounce.ZoomText(self,bs.Actor.RETAINED,
                            str(team.sessionData['score']-1),position=(150,posV),
                            maxWidth=100,
                            shiftPosition=(-100,posV),shiftDelay=shiftDelay,
                            flash=False,
                            trail=False,lifespan=1000,hAlign='left',jitter=1.0)

    def _showTeamScore(self,posV,team,scored,killDelay,shiftDelay):
        bsAnnounce.ZoomText(self,bs.Actor.RETAINED,
                            str(team.sessionData['score']),position=(150,posV),
                            maxWidth=100,
                            shiftPosition=(-100,posV),shiftDelay=shiftDelay,
                            flash=scored,trail=scored,
                            hAlign='left',jitter=1.0,
                            trailColor=(1,0.8,0.0,0))


class TeamSeriesVictoryScoreScreenActivity(TeamsScoreScreenActivity):

    def __init__(self,session,settings={}):
        TeamsScoreScreenActivity.__init__(self,session,settings=settings)
        self._minViewTime = 15000
        self._isFFA = self.getSession().getSessionType() == 'freeForAll'

    def onTransitionIn(self):
        # we dont yet want music and stuff..
        TeamsScoreScreenActivity.onTransitionIn(self,music=None,showTips=False)

    def onBegin(self):
        if bs.getInterfaceType() == 'desktop': s = bs.getResource('pressAnyKeyButtonPlayAgainText')#'Press any key/button to play again...'
        else: s = bs.getResource('pressAnyButtonPlayAgainText')#'Press any button to play again...'

        TeamsScoreScreenActivity.onBegin(self,showUpNext=False,customContinueMessage=s)

        winningTeam = self.settings['winner']

        # pause a moment before playing victory music
        bs.gameTimer(600,bs.WeakCall(self._playVictoryMusic))
        bs.gameTimer(4400,bs.WeakCall(self._showWinner,self.settings['winner']))
        bs.gameTimer(4400+200,bs.WeakCall(bs.playSound,self._scoreDisplaySound))

        if self._isFFA:
            playersSorted = [[p.getTeam().sessionData['score'],name,p] for name,p in self.scoreSet.getValidPlayers().items() if p.getTeam() is not None]
            playersSorted.sort(reverse=True)
        else:
            playersSorted = [[p.score,name,p] for name,p in self.scoreSet.getValidPlayers().items()]
            playersSorted.sort(reverse=True)

        tsHeight = 300
        tsHOffs = -390
        t = 6400
        tIncr = 120

        if self._isFFA:
            txt = bs.getResource('firstToFinalText').replace('${COUNT}',str(self.getSession()._ffaSeriesLength))+':'
        else:
            txt = bs.getResource('bestOfFinalText').replace('${COUNT}',str(self.getSession()._seriesLength))+':'
        bsAnnounce.Text(self,bs.Actor.RETAINED,
                        txt,
                        vAlign='center',
                        maxWidth=300,
                        color=(0.5,0.5,0.5,1.0),position=(0,220),
                        scale=1.2,transition='inTopSlow',hAlign='center',
                        transitionDelay=tIncr*4)

        winScore = (self.getSession()._seriesLength-1)/2+1
        loseScore = 0
        for team in self.teams.values():
            if team.sessionData['score'] != winScore: loseScore = team.sessionData['score']

        if not self._isFFA:
            bsAnnounce.Text(self,bs.Actor.RETAINED,
                            bs.getResource('gamesToText').replace('${WINCOUNT}',str(winScore)).replace('${LOSECOUNT}',str(loseScore)),color=(0.5,0.5,0.5,1.0),
                            maxWidth=160,
                            vAlign='center',
                            position=(0,-215),scale=1.8,transition='inLeft',hAlign='center',
                            transitionDelay=4800+tIncr*4)

        if self._isFFA: vExtra = 120
        else: vExtra = 0

        # show game MVP
        if not self._isFFA:
            mvp,mvpName = None,None
            for p in playersSorted:
                if p[2].getTeam() == winningTeam:
                    mvp = p[2]
                    mvpName = p[1]
                    break
            if mvp is not None:
                bsAnnounce.Text(self,bs.Actor.RETAINED,
                                bs.getResource('mostValuablePlayerText'),color=(0.5,0.5,0.5,1.0),
                                vAlign='center',
                                maxWidth=300,
                                position=(180,tsHeight/2+15),transition='inLeft',
                                hAlign='left',transitionDelay=t)
                t += 4*tIncr

                bsAnnounce.Image(self,bs.Actor.RETAINED,
                                 mvp.getIcon(),position=(238,tsHeight/2-55+17),
                                 scale=(70,70),transition='inLeft',transitionDelay=t)
                bsAnnounce.Text(self,bs.Actor.RETAINED,
                                mvpName,
                                position=(280,tsHeight/2-55+15),hAlign='left',vAlign='center',
                                maxWidth=170,scale=1.3,color=bsUtils.getSafeDisplayColor(mvp.getTeam().color+(1,)),transition='inLeft',
                                transitionDelay=t)
                t += 4*tIncr

        # most violent
        mostKills = 0
        for p in playersSorted:
            if p[2].killCount >= mostKills:
                mvp = p[2]
                mvpName = p[1]
                mostKills = p[2].killCount
        if mvp is not None:
            bsAnnounce.Text(self,bs.Actor.RETAINED,
                            bs.getResource('mostViolentPlayerText'),color=(0.5,0.5,0.5,1.0),
                            vAlign='center',
                            maxWidth=300,
                            position=(180,tsHeight/2-150+vExtra+15),transition='inLeft',
                            hAlign='left',transitionDelay=t)
            bsAnnounce.Text(self,bs.Actor.RETAINED,
                            '('+bs.getResource('killsTallyText').replace('${COUNT}',str(mostKills))+')',
                            position=(260,tsHeight/2-150-15+vExtra),
                            color=(0.3,0.3,0.3,1.0),scale=0.6,hAlign='left',transition='inLeft',
                            transitionDelay=t)
            t += 4*tIncr

            bsAnnounce.Image(self,bs.Actor.RETAINED,
                             mvp.getIcon(),position=(233,tsHeight/2-150-30-46+27+vExtra),scale=(50,50),
                             transition='inLeft',transitionDelay=t)
            bsAnnounce.Text(self,bs.Actor.RETAINED,
                            mvpName,
                            position=(270,tsHeight/2-150-30-36+vExtra+15),hAlign='left',vAlign='center',
                            maxWidth=180,color=bsUtils.getSafeDisplayColor(mvp.getTeam().color+(1,)),transition='inLeft',
                            transitionDelay=t)
            t += 4*tIncr

        # most killed
        mostKilled = 0
        mkp,mkpName = None,None
        for p in playersSorted:
            if p[2].killedCount >= mostKilled:
                mkp = p[2]
                mkpName = p[1]
                mostKilled = p[2].killedCount
        if mkp != None:
            bsAnnounce.Text(self,bs.Actor.RETAINED,
                            bs.getResource('mostViolatedPlayerText'),color=(0.5,0.5,0.5,1.0),
                            vAlign='center',
                            maxWidth=300,
                            position=(180,tsHeight/2-300+vExtra+15),transition='inLeft',
                            hAlign='left',transitionDelay=t)
            bsAnnounce.Text(self,bs.Actor.RETAINED,
                            '('+bs.getResource('deathsTallyText').replace('${COUNT}',str(mostKilled))+')',
                            #"("+str(mostKilled)+" deaths)",
                            position=(260,tsHeight/2-300-15+vExtra),
                            hAlign='left',scale=0.6,color=(0.3,0.3,0.3,1.0),transition='inLeft',
                            transitionDelay=t)
            t += 4*tIncr
            bsAnnounce.Image(self,bs.Actor.RETAINED,
                             mkp.getIcon(),position=(233,tsHeight/2-300-30-46+27+vExtra),
                             scale=(50,50),transition='inLeft',transitionDelay=t)
            bsAnnounce.Text(self,bs.Actor.RETAINED,
                            mkpName,
                            position=(270,tsHeight/2-300-30-36+vExtra+15),
                            hAlign='left',vAlign='center',color=bsUtils.getSafeDisplayColor(mkp.getTeam().color+(1,)),
                            maxWidth=180,transition='inLeft',transitionDelay=t)
            t += 4*tIncr

        # now show individual scores
        tDelay = t
        bsAnnounce.Text(self,bs.Actor.RETAINED,
                        bs.getResource('finalScoresText'),
                        color=(0.5,0.5,0.5,1.0),
                        position=(tsHOffs,tsHeight/2),
                        transition='inRight',
                        transitionDelay=tDelay)
        tDelay += 4*tIncr

        vOffs = 0.0
        tDelay += len(playersSorted) * 8*tIncr
        for score,name,p in playersSorted:
            tDelay -= 4*tIncr
            vOffs -= 40
            bsAnnounce.Text(self,bs.Actor.RETAINED,
                            str(p.getTeam().sessionData['score']) if self._isFFA else str(p.score),
                            color=(0.5,0.5,0.5,1.0),
                            position=(tsHOffs+230,tsHeight/2+vOffs),
                            hAlign='right',transition='inRight',
                            transitionDelay=tDelay)
            tDelay -= 4*tIncr
            
            bsAnnounce.Image(self,bs.Actor.RETAINED,
                             p.getIcon(),position=(tsHOffs-72,tsHeight/2+vOffs+17),
                             scale=(30,30),transition='inLeft',transitionDelay=tDelay)
            bsAnnounce.Text(self,bs.Actor.RETAINED,
                            name,position=(tsHOffs-50,tsHeight/2+vOffs+15),
                            hAlign='left',vAlign='center',maxWidth=180,color=bsUtils.getSafeDisplayColor(p.getTeam().color+(1,)),
                            transition='inRight',transitionDelay=tDelay)

        bs.gameTimer(15000,bs.WeakCall(self._showTips))

    def _showTips(self):
        self._tipsText = bsAnnounce.InfoText(self,bs.Actor.UNRETAINED,offsY=70)

    def _playVictoryMusic(self):
        # make sure we dont stomp on the next activity's music choice
        if not self.isTransitioningOut(): bsUtils.playMusic('Victory')

    def _showWinner(self,team):
        if not self._isFFA: bsAnnounce.ZoomText(self,bs.Actor.RETAINED,team.name,position=(0,97),color=team.color,scale=1.15,jitter=1.0,maxWidth=250)
        else:
            if len(team.players) == 1:
                i = bsAnnounce.Image(self,bs.Actor.RETAINED,
                                 team.players.values()[0].getIcon(),position=(0,130),scale=(80,80))
                bsUtils.animate(i.node,'opacity',{0:0.0,250:1.0})
                t = bsAnnounce.Text(self,bs.Actor.RETAINED,
                                team.players.values()[0].getName(),position=(0,70),scale=1.4,maxWidth=140,
                                hAlign='center',vAlign='center',color=bsUtils.getSafeDisplayColor(team.color+(1,)))
                bsUtils.animate(t.node,'opacity',{0:0.0,250:1.0})

        sExtra = 0.6 if self._isFFA else 1.0
        bsAnnounce.ZoomText(self,bs.Actor.RETAINED,bs.getResource('seriesWinLine1Text'),position=(0,-10),color=team.color,scale=bs.getResource('seriesWinLine1Scale')*sExtra,jitter=1.0,maxWidth=250)
        bsAnnounce.ZoomText(self,bs.Actor.RETAINED,bs.getResource('seriesWinLine2Text'),position=(0,-110),scale=bs.getResource('seriesWinLine2Scale')*sExtra,color=team.color,jitter=1.0,maxWidth=250)


class TeamGameActivity(bs.GameActivity):

    @classmethod
    def getSupportedSessionTypes(self):
        """ return the play modes supported by this game """
        return ['freeForAll','teams']

    def __init__(self,session,level,settings):
        bs.GameActivity.__init__(self,session,level,settings)
        # by default we don't show kill-points in free-for-all
        # (there's usually some activity-specific score and we dont wanna confuse things)
        if session.getSessionType() == 'freeForAll': self._showKillPoints = False

    def onTransitionIn(self,music=None):
        bs.GameActivity.onTransitionIn(self,music)
        self._level = self._levelType(self,bs.Actor.UNRETAINED)

    def onBegin(self,nameSound=None,doStandardPowerups=True):
        bs.GameActivity.onBegin(self,nameSound)
        
        if doStandardPowerups:
            self._powerupDropTimer = bs.Timer(bsPowerup.defaultPowerupInterval,bs.WeakCall(self._dropPowerups),repeat=True)
            self._dropPowerups()
            self._tntObjs = {}
            self._tntDropTimer = bs.Timer(5500,bs.WeakCall(self._dropTnt),repeat=True)
            self._dropTnt()

    def _dropPowerup(self,index,expire=True):
        if self._level != None:
            bsPowerup.Powerup(self,bs.Actor.RETAINED,translate=self._level.powerupSpawnPoints[index],
                              powerupType=bsPowerup.getFactory().getRandomPowerupType(),expire=expire)

    def _dropPowerups(self):
        """ Standard powerup drop """
        # drop one powerup per point
        pts = self._level.powerupSpawnPoints
        for i,pt in enumerate(pts):
            bs.gameTimer(i*400,bs.WeakCall(self._dropPowerup,i))

    def _dropTnt(self):
        """ Standard tnt drop """
        # drop tnt on the map for any tnt location with no existing tnt box
        for i,pt in enumerate(self._level.tntPoints):
            if not i in self._tntObjs: self._tntObjs[i] = {'absentTicks':9999,'obj':None}
            tntObj = self._tntObjs[i]

            # respawn once its been dead for a while..
            if tntObj['obj'] is None or not tntObj['obj'].exists():
                tntObj['absentTicks'] += 1
                if tntObj['absentTicks'] > 3:
                    tntObj['obj'] = bsBomb.Bomb(self,bs.Actor.UNRETAINED,
                                                translate=pt,bombType='tnt')
                    tntObj['absentTicks'] = 0

    def spawnSpazForPlayer(self,player,position=None,angle=None):
        "Spawn and wire up a standard player spaz"
        if position is None:
            # in teams-mode get our team-start-location
            if self.getSession().getSessionType() == 'teams':
                position = self._level.getStartPosition(player.getTeam().teamID)
            else:
                # otherwise do free-for-all spawn locations
                position = self._level.getFFAStartPosition(self.players.values())
            
        return bs.GameActivity.spawnSpazForPlayer(self,player,position,angle)

    def end(self,results=None,announceWinningTeam=True,announceDelay=100):
        # announce win (but just for the first end() call)
        if not self.hasEnded():
            self.getSession().announceGameResults(self,results,delay=announceDelay,announceWinningTeam=announceWinningTeam)
        bs.GameActivity.end(self,results,delay=2000+announceDelay)
            

class TeamGameSession(bs.Session):

    def __init__(self):

        mp = self.getMaxPlayers()

        bs.Session.__init__(self,teamNames=gTeamNames if self._useTeams else None,
                            teamColors=gTeamColors if self._useTeams else None,
                            useTeamColors=self._useTeams,
                            maxPlayers=mp)

        self._seriesLength = 7
        self._ffaSeriesLength = 24

        try: self._playlistName = bs.config[self._playlistSelectionVar]
        except Exception: self._playlistName = self._defaultPlaylistName
        try: self._playlistRandomize = bs.config[self._playlistRandomizeVar]
        except Exception: self._playlistRandomize = False

        self._roundEndSound = bs.getSound("boxingBell")
        self._redWinsSound = bs.getSound('announceRedWins')
        self._blueWinsSound = bs.getSound('announceBlueWins')

        # which game activity we're on
        self._gameNumber = 0

        try: playlists = bs.config[self._playlistsVar]
        except Exception: playlists = {}

        if self._playlistName != self._defaultPlaylistName and self._playlistName in playlists:
            # (make sure to copy this, as we muck with it in place once we've got it and we dont want that to affect our config)
            playlist = copy.deepcopy(playlists[self._playlistName])
        else:
            if self._useTeams: playlist = bs.getDefaultTeamsPlaylist()
            else: playlist = bs.getDefaultFreeForAllPlaylist()

        # resolve types and whatnot to get our final playlist
        playlistResolved = bs.filterPlaylist(playlist,sessionType=self.getSessionType(),addResolvedType=True)

        self._playlist = ShuffleList(playlistResolved,shuffle=self._playlistRandomize)

        # get a game on deck ready to go
        self._currentGameSpec = None
        self._nextGameSpec = self._playlist.pullNext()
        self._nextGame = self._nextGameSpec['resolvedType']

        # go ahead and instantiate the next game we'll use so it has lots of time to load
        self._instantiateNextGame()

        # start in standard join screen
        j = bs.JoiningActivity(self)
        self.setActivity(j)

        # show info about the next up game
        j._nextUpText = bsAnnounce.Text(j,bs.Actor.UNRETAINED,
                                        bs.getResource('upFirstText')+'  '+self.getNextGameName(),
                                        hAttach='center',
                                        scale=0.7,
                                        vAttach='top',
                                        hAlign='center',
                                        position=(0,-70),
                                        flash=False,
                                        color=(0.5,0.5,0.5,1.0),
                                        transition='fadeIn',
                                        transitionDelay=5000)

    def getNextGameName(self):
        return self._nextGameSpec['resolvedType'].getConfigDescription(self._nextGameSpec)

    def getGameNumber(self):
        return self._gameNumber

    def onTeamJoined(self,team):
        team.sessionData['previousScore'] = team.sessionData['score'] = 0

    def getSessionType(self):
        return 'teams' if self._useTeams else 'freeForAll'

    def getMaxPlayers(self):
        if self._useTeams:
            try: return bs.config['Team Game Max Players']
            except Exception: return 8
        else:
            try: return bs.config['Free-for-All Max Players']
            except Exception: return 8
            

    def _instantiateNextGame(self):
        self._nextGameInstance = self._nextGameSpec['resolvedType'](self,self._nextGameSpec['map'],self._nextGameSpec['settings'])

    def onPlayerRequest(self,playerID):
        return bs.Session.onPlayerRequest(self,playerID)

    def getFFAPointAwards(self):
        'return the number of points awarded based on placement for the current FFA game'
        if len(self.players) == 1: pointAwards = {}
        elif len(self.players) == 2: pointAwards = {0:6}
        elif len(self.players) == 3: pointAwards = {0:6, 1:3}
        elif len(self.players) == 4: pointAwards = {0:8, 1:4, 2:2}
        elif len(self.players) == 5: pointAwards = {0:8, 1:4, 2:2}
        elif len(self.players) == 6: pointAwards = {0:8, 1:4, 2:2}
        else: pointAwards = {0:8, 1:4, 2:2, 3:1}
        return pointAwards

    def onActivityEnd(self,activity,results):

        # if we're in a between-round activity or a restart-activity, hop into a round
        if (isinstance(activity,bs.JoiningActivity)
            or isinstance(activity,bs.ScoreScreenActivity)):

            # if we're coming from a series-end activity, reset scores
            if isinstance(activity,TeamSeriesVictoryScoreScreenActivity):
                self.scoreSet.reset()
                self._gameNumber = 0
                for team in self.teams.values(): team.sessionData['score'] = 0
            # otherwise just set accum (per-game) scores
            else: self.scoreSet.resetAccum()

            nextGame = self._nextGameInstance

            self._currentGameSpec = self._nextGameSpec
            self._nextGameSpec = self._playlist.pullNext()
            self._gameNumber += 1

            # instantiate the next now so they have plenty of time to load
            self._instantiateNextGame()

            # (re)register all players and wire the score-set to our next activity
            for p in self.players.values(): self.scoreSet.registerPlayer(p)
            self.scoreSet.setActivity(nextGame)

            # now flip the current activity
            self.setActivity(nextGame)

        # if we're leaving a round, go to the score screen
        else:
            # looks like we were in a round - go to the score screen
            if self._useTeams:
                winners = results.getWinners()
                # if everyone has the same score, call it a draw
                if len(winners) < 2: self.setActivity(DrawScoreScreenActivity(self,{}))
                else:
                    winner = winners[0]['teams'][0]
                    winner.sessionData['score'] += 1
                    # if a team has won, show final victory screen...
                    if winner.sessionData['score'] >= (self._seriesLength-1)/2+1: self.setActivity(TeamSeriesVictoryScoreScreenActivity(self,{'winner':winner}))
                    else: self.setActivity(TeamVictoryScoreScreenActivity(self,{'winner':winner}))
            else:
                winners = results.getWinners()

                # if there's multiple players and everyone has the same score, call it a draw
                if len(self.players) > 1 and len(winners) < 2: self.setActivity(DrawScoreScreenActivity(self,{'results':results}))
                else:
                    # award different point amounts based on number of players
                    pointAwards = self.getFFAPointAwards()

                    for i,winner in enumerate(winners):
                        for team in winner['teams']:
                            points = pointAwards[i] if i in pointAwards else 0
                            team.sessionData['previousScore'] = team.sessionData['score']
                            team.sessionData['score'] += points

                    seriesWinners = [team for team in self.teams.values() if team.sessionData['score'] >= self._ffaSeriesLength]
                    seriesWinners.sort(reverse=True,key=lambda team:(team.sessionData['score']))
                    if len(seriesWinners) == 1 or (len(seriesWinners) > 1 and seriesWinners[0].sessionData['score'] != seriesWinners[1].sessionData['score']):
                        self.setActivity(TeamSeriesVictoryScoreScreenActivity(self,{'winner':seriesWinners[0]}))
                    else: self.setActivity(FreeForAllVictoryScoreScreenActivity(self,{'results':results}))

    def announceGameResults(self,activity,results,delay,announceWinningTeam=True):

        bs.gameTimer(delay,bs.WeakCall(bs.playSound,self._roundEndSound))
        if announceWinningTeam:
            winningTeam = results.getWinningTeam()
            if winningTeam is not None:
                # have all players celebrate
                for player in winningTeam.players.values():
                    try: bs.nodeMessage(player.actor.node,'celebrate',10000)
                    except Exception: pass
                activity.cameraFlash()

                # announce winner
                if self._useTeams:
                    if winningTeam.name == bs.GameActivity.getTeamNameTranslated('Red'): bs.gameTimer(delay+300,bs.WeakCall(bs.playSound,self._redWinsSound))
                    elif winningTeam.name == bs.GameActivity.getTeamNameTranslated('Blue'): bs.gameTimer(delay+300,bs.WeakCall(bs.playSound,self._blueWinsSound))

                activity.showZoomMessage(bs.getResource('winsText').replace('${NAME}',winningTeam.name),scale=0.85,color=bsUtils.getNormalizedColor(winningTeam.color))

                
class FreeForAllSession(TeamGameSession):
    _useTeams = False
    _playlistSelectionVar = 'Free-for-All Playlist Selection'
    _defaultPlaylistName = 'Default Free-for-All Games' # FIXME need to use translated value
    _playlistRandomizeVar = 'Free-for-All Playlist Randomize'
    _playlistsVar = 'Free-for-All Playlists'
    
class TeamsSession(TeamGameSession):
    _useTeams = True
    _playlistSelectionVar = 'Team Tournament Playlist Selection'
    _defaultPlaylistName = 'Default Team Games' # FIXME need to use translated value
    _playlistRandomizeVar = 'Team Tournament Playlist Randomize'
    _playlistsVar = 'Team Tournament Playlists'

